from __future__ import annotations

from typing import Any, Dict, Sequence, Type, Union

import web3
from web3 import eth, geth, module, net

from huma_signals.commons import chains

_WEB3_CHAIN_NETWORK_ID: Dict[chains.Chain, str] = {
    chains.Chain.ETHEREUM: "1",
    chains.Chain.GOERLI: "5",
    chains.Chain.POLYGON: "137",
    chains.Chain.MUMBAI: "80001",
}

_MODULES: Dict[str, Union[Type[module.Module], Sequence[Any]]] = {
    "eth": eth.AsyncEth,  # type: ignore[attr-defined]
    "net": net.AsyncNet,
    "geth": (
        geth.Geth,
        {
            "txpool": geth.AsyncGethTxPool,
            "personal": geth.AsyncGethPersonal,
            "admin": geth.AsyncGethAdmin,
        },
    ),
}


async def get_w3(chain: chains.Chain, web3_provider_url: str) -> web3.Web3:
    w3 = web3.Web3(
        provider=web3.Web3.AsyncHTTPProvider(web3_provider_url), modules=_MODULES
    )
    if await w3.net.version != _WEB3_CHAIN_NETWORK_ID.get(chain):  # type: ignore
        raise ValueError(f"Web3 provider is not compatible with chain {chain.name}")
    return w3



import datetime
import decimal

import pydantic

from huma_signals import models


class Invoice(models.HumaBaseModel):
    token_owner: str = pydantic.Field(description="The address of the token owner")
    currency: str = pydantic.Field(
        description="The currency of the invoice (e.g. USDC, ETH, etc.)"
    )
    amount: decimal.Decimal = pydantic.Field(
        description="The amount of the invoice (e.g. 100)"
    )
    status: str = pydantic.Field(
        description="The status of the invoice (e.g. PAID, UNPAID, etc.)"
    )
    payer: str = pydantic.Field(description="The payer's address")
    payee: str = pydantic.Field(description="The payee's address")
    creation_date: datetime.datetime = pydantic.Field(
        description="The date the invoice was created"
    )
    due_date: datetime.datetime = pydantic.Field(
        description="The date the invoice is due"
    )



from typing import Any, ClassVar, List

import httpx

from huma_signals import models
from huma_signals.commons import chains
from huma_signals.domain.adapters import models as adapter_models
from huma_signals.settings import settings


class AllowListSignal(models.HumaBaseModel):
    """
    Signals emitted by the allowlist adapter.

    Note this is a temporary adapter and will be removed once the allowlist is integrated into the EAs.
    """

    on_allowlist: bool = False


class AllowListAdapter(adapter_models.SignalAdapterBase):
    name: ClassVar[str] = "allowlist"
    required_inputs: ClassVar[List[str]] = ["borrower_wallet_address"]
    signals: ClassVar[List[str]] = list(AllowListSignal.__fields__.keys())

    def __init__(
        self,
        chain: chains.Chain = settings.chain,
        allowlist_endpoint: str = settings.allow_list_endpoint,
    ) -> None:
        if not allowlist_endpoint:
            raise ValueError("allowlist_endpoint is empty")

        self.chain = chain
        self.allowlist_endpoint = allowlist_endpoint

    async def fetch(  # pylint: disable=arguments-differ
        self,
        borrower_wallet_address: str,
        *args: Any,
        **kwargs: Any,
    ) -> AllowListSignal:
        url = f"/user/{'testnet' if self.chain.is_testnet() else 'mainnet'}/{borrower_wallet_address}"
        try:
            async with httpx.AsyncClient(base_url=self.allowlist_endpoint) as client:
                resp = await client.get(url=url)
                resp.raise_for_status()
                if resp.json().get("status") == "found":
                    return AllowListSignal(on_allowlist=True)
        except httpx.HTTPStatusError:
            pass

        return AllowListSignal(on_allowlist=False)



# pylint: disable=too-few-public-methods
from typing import Any

from huma_signals import models


class SignalFetchRequest(models.HumaBaseModel):
    signal_names: list[str]
    adapter_inputs: dict[str, Any]


class SignalFetchResponse(models.HumaBaseModel):
    signals: dict[str, Any]



from typing import Protocol

import httpx
import structlog

from huma_signals.domain.clients.polygon_client import polygon_types

logger = structlog.get_logger(__name__)


class BasePolygonClient(Protocol):
    async def get_transactions(
        self, wallet_address: str
    ) -> list[polygon_types.PolygonTransaction]:
        pass


class PolygonClient(BasePolygonClient):
    def __init__(
        self,
        polygonscan_base_url: str,
        polygonscan_api_key: str,
    ) -> None:
        self.polygonscan_base_url = polygonscan_base_url
        self.polygonscan_api_key = polygonscan_api_key

    async def get_transactions(
        self, wallet_address: str
    ) -> list[polygon_types.PolygonTransaction]:
        try:
            async with httpx.AsyncClient(base_url=self.polygonscan_base_url) as client:
                request = (
                    f"/api?module=account&action=txlist"
                    f"&address={wallet_address}"
                    f"&startblock=0&endblock=99999999"
                    f"&sort=asc"
                    f"&apikey={self.polygonscan_api_key}"
                )
                resp = await client.get(request)

                resp.raise_for_status()
                payload = polygon_types.PolygonTransactionResponse(**resp.json())
                if payload.status == "1":
                    return payload.result
        except httpx.HTTPStatusError:
            logger.exception("Error fetching transactions", request=request)

        return []



import datadog
import fastapi
import structlog
from ddtrace import opentracer
from fastapi.middleware import cors

from huma_signals.api import views
from huma_signals.settings import settings

logger = structlog.get_logger(__name__)

_DATADOG_HOST = "0.0.0.0"

if settings.instrumentation_enabled:
    datadog.initialize(
        api_key=settings.datadog_api_key,
        statsd_host=_DATADOG_HOST,
        statsd_port=8125,
    )
    tracer = opentracer.Tracer(
        "",
        config={
            "agent_hostname": _DATADOG_HOST,
            "agent_port": 8126,
        },
    )
    opentracer.set_global_tracer(tracer)

app = fastapi.FastAPI()

# for javascript to call the endpoints
origins = ["*"]
app.add_middleware(
    cors.CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.include_router(views.router, prefix="", tags=["huma_signals"])


@app.get("/health")
async def get_health() -> str:
    return "ok"



import pydantic

from huma_signals.commons import pydantic_utils


class EthTransaction(pydantic_utils.CamelCaseAliased):
    block_number: str
    time_stamp: str
    hash: str
    nonce: str
    block_hash: str
    transaction_index: str
    from_: str = pydantic.Field(alias="from")
    to: str
    value: str
    gas: str
    gas_price: str
    is_error: str
    txreceipt_status: str
    input: str
    contract_address: str
    cumulative_gas_used: str
    gas_used: str
    confirmations: str
    method_id: str
    function_name: str


class EthTransactionResponse(pydantic_utils.CamelCaseAliased):
    status: str
    message: str
    result: list[EthTransaction]



from typing import Any, Type

from huma_signals.domain.adapters import models
from huma_signals.domain.adapters.allowlist import adapter as allowlist_adapter
from huma_signals.domain.adapters.ethereum_wallet import (
    adapter as ethereum_wallet_adapter,
)
from huma_signals.domain.adapters.lending_pools import adapter as lending_pools_adapter
from huma_signals.domain.adapters.polygon_wallet import (
    adapter as polygon_wallet_adapter,
)
from huma_signals.domain.adapters.request_network import (
    request_invoice_adapter,
    request_transaction_adapter,
)

ADAPTER_REGISTRY: dict[str, Type[models.SignalAdapterBase]] = {
    lending_pools_adapter.LendingPoolAdapter.name: lending_pools_adapter.LendingPoolAdapter,
    request_invoice_adapter.RequestInvoiceAdapter.name: request_invoice_adapter.RequestInvoiceAdapter,
    request_transaction_adapter.RequestTransactionAdapter.name: request_transaction_adapter.RequestTransactionAdapter,
    allowlist_adapter.AllowListAdapter.name: allowlist_adapter.AllowListAdapter,
    ethereum_wallet_adapter.EthereumWalletAdapter.name: ethereum_wallet_adapter.EthereumWalletAdapter,
    polygon_wallet_adapter.PolygonWalletAdapter.name: polygon_wallet_adapter.PolygonWalletAdapter,
}


def find_required_adapter(
    signal_names: list[str],
    adapter_registry: dict[str, Type[models.SignalAdapterBase]] | None = None,
) -> list[Type[models.SignalAdapterBase]]:
    """Find the adapter required to fetch the signals"""
    if adapter_registry is None:
        adapter_registry = ADAPTER_REGISTRY

    adapters = set()
    for signal_name in signal_names:
        adapter_name = signal_name.split(".")[0]
        adapter = adapter_registry.get(adapter_name)
        if adapter is None:
            raise KeyError(f"Signal Adapter {adapter_name} not found in registry")
        if signal_name.split(".")[1] not in adapter.signals:
            raise KeyError(f"Signal {signal_name} not found in adapter {adapter_name}")
        adapters.add(adapter_registry[adapter_name])

    return list(adapters)


async def fetch_signal(
    signal_names: list[str],
    adapter_inputs: dict[str, Any],
    registry: dict[str, Type[models.SignalAdapterBase]] | None = None,
) -> dict[str, Any]:
    """Fetch signals from the signal adapters"""
    if registry is None:
        registry = ADAPTER_REGISTRY

    adapters = find_required_adapter(signal_names, registry)
    all_signals: dict[str, Any] = {}
    for adapter in adapters:
        inputs = {k: adapter_inputs[k] for k in adapter.required_inputs}
        signals = (await adapter().fetch(**inputs)).dict()
        all_signals |= {
            f"{adapter.name}.{signal}": value for signal, value in signals.items()
        }
    return {k: v for k, v in all_signals.items() if k in signal_names}



from typing import Any, Type

import fastapi
import structlog

from huma_signals.api import exception_handlers, models
from huma_signals.domain.adapters import models as adapter_models
from huma_signals.domain.adapters import registry

logger = structlog.get_logger(__name__)

router = fastapi.APIRouter()


def _list_adapters(
    registry_: dict[str, Type[adapter_models.SignalAdapterBase]]
) -> dict[str, Any]:
    return {"adapters": [adapter.definition() for adapter in registry_.values()]}


@router.get("/list_adapters")
@exception_handlers.handle_exception
async def get_list_adapters() -> dict[str, Any]:
    logger.info("Received list adapters request")
    return _list_adapters(registry.ADAPTER_REGISTRY)


@router.post("/fetch")
@exception_handlers.handle_exception
async def fetch(
    signal_request: models.SignalFetchRequest,
) -> models.SignalFetchResponse:
    logger.info("Received signal request", request=signal_request)
    signals = await registry.fetch_signal(
        signal_request.signal_names, signal_request.adapter_inputs
    )
    return models.SignalFetchResponse(signals=signals)



import datetime
from typing import Any, ClassVar

import structlog

from huma_signals import models
from huma_signals.commons import datetime_utils
from huma_signals.domain.adapters import models as adapter_models
from huma_signals.domain.adapters.polygon_wallet.settings import settings
from huma_signals.domain.clients.polygon_client import polygon_client, polygon_types

logger = structlog.get_logger()


class PolygonWalletSignals(models.HumaBaseModel):
    total_transactions: int
    total_sent: int
    total_received: int
    wallet_tenure_in_days: int
    total_income_90days: float
    total_transactions_90days: int


class BasePolygonWalletAdapter(adapter_models.SignalAdapterBase):
    name: ClassVar[str] = "polygon_wallet"
    required_inputs: ClassVar[list[str]] = ["borrower_wallet_address"]
    signals: ClassVar[list[str]] = list(PolygonWalletSignals.__fields__.keys())

    async def fetch(  # pylint: disable=arguments-differ
        self, borrower_wallet_address: str, *args: Any, **kwargs: Any
    ) -> PolygonWalletSignals:
        raise NotImplementedError


class PolygonWalletAdapter(BasePolygonWalletAdapter):
    def __init__(
        self,
        polygon_client_: polygon_client.BasePolygonClient | None = None,
        polygonscan_base_url: str = settings.polygonscan_base_url,
        polygonscan_api_key: str = settings.polygonscan_api_key,
    ) -> None:
        self.polygon_client = polygon_client_ or polygon_client.PolygonClient(
            polygonscan_base_url=polygonscan_base_url,
            polygonscan_api_key=polygonscan_api_key,
        )

    async def fetch(
        self, borrower_wallet_address: str, *args: Any, **kwargs: Any
    ) -> PolygonWalletSignals:
        transactions = await self.polygon_client.get_transactions(
            borrower_wallet_address,
        )
        now = datetime_utils.tz_aware_utc_now()
        min_tx_timestamp = (
            None
            if len(transactions) == 0
            else min(tx.time_stamp for tx in transactions)
        )

        return PolygonWalletSignals(
            total_transactions=len(transactions),
            total_sent=len(
                [
                    tx
                    for tx in transactions
                    if tx.from_ == borrower_wallet_address.lower()
                ]
            ),
            total_received=len(
                [tx for tx in transactions if tx.to == borrower_wallet_address.lower()]
            ),
            wallet_tenure_in_days=0
            if min_tx_timestamp is None
            else (
                now
                - datetime_utils.timestamp_to_tz_aware_utc_datetime(min_tx_timestamp)
            ).days,
            total_income_90days=sum(
                int(tx.value)
                for tx in transactions
                if _is_transaction_within_90_days(now=now, transaction=tx)
                and tx.to == borrower_wallet_address.lower()
            ),
            total_transactions_90days=len(
                [
                    tx
                    for tx in transactions
                    if _is_transaction_within_90_days(now=now, transaction=tx)
                ]
            ),
        )


def _is_transaction_within_90_days(
    now: datetime.datetime, transaction: polygon_types.PolygonTransaction
) -> bool:
    return (
        now - datetime_utils.timestamp_to_tz_aware_utc_datetime(transaction.time_stamp)
    ).days < 90



from __future__ import annotations

import enum


class Chain(enum.Enum):
    ETHEREUM = "ETHEREUM"
    GOERLI = "GOERLI"
    POLYGON = "POLYGON"
    MUMBAI = "MUMBAI"

    @staticmethod
    def from_chain_name(chain_name: str) -> Chain:
        if chain_name.lower() in ("ethereum", "mainnet", "eth", "homestead"):
            return Chain.ETHEREUM
        if chain_name.lower() in ("goerli"):
            return Chain.GOERLI
        if chain_name.lower() in ("polygon", "matic"):
            return Chain.POLYGON
        if chain_name.lower() in ("mumbai"):
            return Chain.MUMBAI
        raise ValueError(f"Unsupported chain: {chain_name}")

    def chain_name(self) -> str:
        return self.name.lower()

    def is_testnet(self) -> bool:
        return self.chain_name() in ("goerli", "mumbai")



import functools
from typing import Any, Callable, Type

import fastapi
import structlog

from huma_signals import exceptions

logger = structlog.get_logger(__name__)


def handle_exception(f: Callable) -> Callable:
    @functools.wraps(f)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return await f(*args, **kwargs)
        except exceptions.HumaSignalException as e:
            logger.exception(f"Error calling {f}", error_message=e.message)
            status_code = _EXCEPTIONS_TO_HTTP_STATUS_CODES.get(type(e), 500)
            raise fastapi.HTTPException(
                status_code=status_code,
                detail={"type": e.__class__.__name__, "message": e.message},
            )
        except Exception as e:
            logger.exception(f"Error calling {f}", e=e)
            raise fastapi.HTTPException(
                status_code=500,
                detail={"type": e.__class__.__name__, "message": str(e)},
            )

    return wrapper


_EXCEPTIONS_TO_HTTP_STATUS_CODES: dict[Type[exceptions.HumaSignalException], int] = {
    exceptions.InvalidAddressException: 400,
    exceptions.PoolSettingsNotFoundException: 404,
}



from typing import Any, ClassVar


class SignalAdapterBase:
    # Signal adapters name.
    name: ClassVar[str]
    # Signals that the adapter can fetch.
    signals: ClassVar[list[str]]
    # Inputs that the adapter requires.
    required_inputs: ClassVar[list[str]]

    @classmethod
    def definition(cls) -> dict[str, str | list[str]]:
        return {
            "name": cls.name,
            "signals": cls.signals,
            "required_inputs": cls.required_inputs,
        }

    async def fetch(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError
